/// <reference path="../typings/index.d.ts" />
"use strict";
var path = require('path');
var fs = require('fs');
var copy_file_1 = require('./lib/copy-file');
var find_files_1 = require('./lib/find-files');
var patterns = require('./lib/regexp-patterns');
/**
 * Global verbose on/off switch. Import and set to `false` to remove
 * all debug output except for Exceptions.
 * @type {boolean}
 */
exports.DTSBuilderVerbose = true;
/**
 * Receives a list of Bundles and generates the d.ts file according
 * to bundle configuration.
 * @param {Array<Bundle>} bundles
 */
function generateBundles(bundles) {
    if (!Array.isArray(bundles)) {
        throw new TypeError("Invalid bundles argument, must be an array. Got " + typeof bundles + ".");
    }
    var len = bundles.length;
    log('Generating definition for %d bundle(s)...', len);
    if (!len) {
        log('Nothing to do... Bye.');
        return;
    }
    bundles.forEach(function (bundle) {
        log('> Generating bundle %s', bundle.name);
        log(' + Gathering definitions from %s/**/*.d.ts', bundle.sourceDir);
        var dtsFiles = find_files_1.findFiles(bundle.sourceDir);
        if (!dtsFiles.length) {
            log('No files were found in bundle %s.', bundle.name);
            return;
        }
        log('   + Found %d definition files.', dtsFiles.length);
        log(' + Concatenating definitions to a single file');
        var binBuffers = dtsFiles.map(function (file) { return fs.readFileSync(file); });
        var txtBuffer = binBuffers.map(function (buffer) { return buffer.toString(); }).join('\n');
        log('   + Done, total size: %d characters', txtBuffer.length);
        log(' + Optimizing imports...');
        txtBuffer = optimizeImports(txtBuffer);
        log(' + Wrapping with module...');
        txtBuffer = moduleWrap(txtBuffer, bundle);
        log(' + Removing empty lines...');
        txtBuffer = cleanEmptyLines(txtBuffer);
        if (bundle.externals) {
            log(' + Adding external definition files...');
            txtBuffer = addExternalRefs(bundle) + '\n' + txtBuffer;
        }
        var destFile = pathAppend(bundle.destDir, toKebab(bundle.name) + ".d.ts");
        log(' +-> Saving %s...', destFile);
        fs.writeFile(destFile, txtBuffer, function (err) {
            if (err) {
                throw new Error(err.message);
            }
            log('! "%s" bundle is ready', bundle.name);
        });
    });
}
exports.generateBundles = generateBundles;
/*****************************************************************************/
function toKebab(name) {
    if (!name)
        return null;
    return name.split(/\W+/).map(function (part) { return part.toLowerCase(); }).join('-');
}
function toCamel(name) {
    if (!name)
        return null;
    var parts = name.split(/\W+/).map(function (part) { return part.toLowerCase(); });
    for (var i = 1; i < parts.length; i++) {
        parts[i] = parts[i].charAt(0).toUpperCase() + parts[i].substr(1);
    }
    return parts.join('');
}
/**
 * @param {string} src
 * @param {string} dst
 * @returns {string}
 */
function pathAppend(src, dst) {
    src = src.trim().replace(/\/+$/, '');
    dst = dst.trim().replace(/^\/+/, '');
    return [src, dst].join('/');
}
/**
 * @param {string} txtBuffer
 * @returns {string}
 */
function optimizeImports(txtBuffer) {
    // remove imports
    var result = txtBuffer.replace(patterns.externalModule, '');
    // get unique internal imports
    var matches = result.match(patterns.internalModule);
    var internalImports = matches ? getUnique(matches.map(function (imp) { return imp.toString().trim(); })) : [];
    // remove internal imports
    result = result.replace(patterns.internalModule, '');
    var remaining = result.match(/\bimport\b/);
    if (remaining && remaining.length) {
        console.error('Could not deal with the following:', JSON.stringify(remaining, null, 2));
    }
    // add internals
    result = sortInternals(internalImports).join('\n') + '\n' + result;
    result = removeReferences(result);
    result = removeDefaults(result);
    // remove export * from ...
    result = result.replace(patterns.externalReExports, '');
    // remove export declarations
    result = result.replace(patterns.exportDeclarations, '');
    return result;
}
/**
 * Returns unique array elements
 * @param {Array} arr
 * @returns {Array}
 */
function getUnique(arr) {
    var res = [];
    arr.forEach(function (item) {
        if (res.indexOf(item) === -1)
            res.push(item);
    });
    return res;
}
/**
 * @param {Array} arr
 * @returns {Array}
 */
function sortInternals(arr) {
    return arr
        .map(function (line) { return line.match(patterns.internalModuleParts); })
        .sort(sortMatches)
        .map(function (obj) { return ("import " + obj[1] + " = " + obj[2] + ";"); });
}
/**
 * @param {Array} a
 * @param {Array} b
 * @returns {number}
 * @private
 */
function sortMatches(a, b) {
    var impA = a[2].toLowerCase();
    var impB = b[2].toLowerCase();
    if (impA < impB)
        return -1;
    if (impA > impB)
        return +1;
    return 0;
}
/**
 * @param {string} text
 * @returns {string}
 */
function removeReferences(text) {
    return text.replace(patterns.reference, '');
}
/**
 * @param {Bundle} bundle
 * @returns {string}
 */
function addExternalRefs(bundle) {
    var res = bundle.externals.map(function (ref) {
        var fileName = path.basename(ref);
        copy_file_1.copyFile(ref, pathAppend(bundle.destDir, fileName), function (err, res) {
            if (err) {
                throw new Error(err.message);
            }
            // TODO: localize references of the copied file, or update path
            log('+++ Copied external reference: %s', path.basename(res));
        });
        log(' * Referencing external: %s', fileName);
        return "/// <reference path=\"" + fileName + "\" />";
    });
    return res.join('\n');
}
/**
 * @param {string} text
 * @returns {string}
 */
function removeDefaults(text) {
    return text
        .replace(patterns.defaultExport, '')
        .replace(patterns.defaultsFromNs, '');
}
/**
 * @param {string} text
 * @param {Bundle} bundle
 * @returns {string}
 */
function moduleWrap(text, bundle) {
    log(' * Converting to module...');
    var camelName = toCamel(bundle.name);
    var lines = text
        .split('\n')
        .map(function (line) { return '  ' + line
        .replace(/\bexport\s+declare\s+/, 'export ')
        .replace(/\bdeclare\s+(class|function|const|var|let)\s+/, '$1 '); });
    lines.unshift("\ndeclare module '" + camelName + "' {\n  export = " + camelName + ";\n}\n\ndeclare namespace " + camelName + " {");
    lines.push('}');
    // add alias
    if (bundle.alias) {
        lines.unshift("import " + bundle.alias + " = " + camelName + ";");
    }
    return lines.join('\n');
}
/**
 * Remove empty lines, when 2 or more in a sequence
 * @param {string} text
 * @returns {string}
 */
function cleanEmptyLines(text) {
    var prevEmpty = false;
    return text.split('\n').filter(function (line) {
        var isEmpty = !line.trim().length;
        var keep = !isEmpty || !prevEmpty;
        prevEmpty = isEmpty;
        return keep;
    }).join('\n');
}
/**
 * Logging helper
 */
function log(message) {
    var args = [];
    for (var _i = 1; _i < arguments.length; _i++) {
        args[_i - 1] = arguments[_i];
    }
    if (!exports.DTSBuilderVerbose)
        return;
    console.log.apply(console, [message].concat(args));
}
//# sourceMappingURL=index.js.map